const sext = @import("zba-util").sext;

pub fn fmt16(comptime InstrFn: type, comptime cond: u4) InstrFn {
    const Arm32 = @typeInfo(@typeInfo(@typeInfo(InstrFn).Pointer.child).Fn.params[0].type.?).Pointer.child;

    return struct {
        fn inner(cpu: *Arm32, opcode: u16) void {
            // B
            if (cond == 0xE or cond == 0xF)
                cpu.panic("[CPU/THUMB.16] Undefined conditional branch with condition {}", .{cond});

            if (!cpu.cpsr.check(Arm32.arch, cond)) return;

            cpu.r[15] +%= sext(u32, u8, opcode & 0xFF) << 1;
            cpu.pipe.reload(cpu);
        }
    }.inner;
}

pub fn linkExchange(comptime InstrFn: type, comptime H: u2) InstrFn {
    const Arm32 = @typeInfo(@typeInfo(@typeInfo(InstrFn).Pointer.child).Fn.params[0].type.?).Pointer.child;

    return struct {
        fn inner(cpu: *Arm32, opcode: u16) void {
            const offset = opcode & 0x7FF;

            switch (H) {
                0b00 => { // Unconditional Branch
                    cpu.r[15] +%= sext(u32, u11, opcode & 0x7FF) << 1;
                    cpu.pipe.reload(cpu);
                },
                0b01 => { // BLX Pt. 2
                    if (Arm32.arch == .v4t) cpu.panic("attempted to execute THUMB BLX(1), despite ARMv4T CPU", .{});
                    const next_addr = cpu.r[15] - 2;

                    cpu.r[15] = (cpu.r[14] +% (offset << 1)) & ~@as(u32, 0x3);
                    cpu.r[14] = next_addr | 1;
                    cpu.cpsr.t.unset();

                    cpu.pipe.reload(cpu);
                },
                0b10 => cpu.r[14] = cpu.r[15] +% (sext(u32, u11, offset) << 12), // BL / BLX Pt. 1
                0b11 => { // BL Pt. 2
                    const next_addr = cpu.r[15] - 2;

                    cpu.r[15] = cpu.r[14] +% (offset << 1);
                    cpu.r[14] = next_addr | 1;

                    cpu.pipe.reload(cpu);
                },
            }
        }
    }.inner;
}
